# bash completion for flightctl                            -*- shell-script -*-

__flightctl_debug()
{
    if [[ -n ${BASH_COMP_DEBUG_FILE:-} ]]; then
        echo "$*" >> "${BASH_COMP_DEBUG_FILE}"
    fi
}

# Homebrew on Macs have version 1.3 of bash-completion which doesn't include
# _init_completion. This is a very minimal version of that function.
__flightctl_init_completion()
{
    COMPREPLY=()
    _get_comp_words_by_ref "$@" cur prev words cword
}

__flightctl_index_of_word()
{
    local w word=$1
    shift
    index=0
    for w in "$@"; do
        [[ $w = "$word" ]] && return
        index=$((index+1))
    done
    index=-1
}

__flightctl_contains_word()
{
    local w word=$1; shift
    for w in "$@"; do
        [[ $w = "$word" ]] && return
    done
    return 1
}

__flightctl_handle_go_custom_completion()
{
    __flightctl_debug "${FUNCNAME[0]}: cur is ${cur}, words[*] is ${words[*]}, #words[@] is ${#words[@]}"

    local shellCompDirectiveError=1
    local shellCompDirectiveNoSpace=2
    local shellCompDirectiveNoFileComp=4
    local shellCompDirectiveFilterFileExt=8
    local shellCompDirectiveFilterDirs=16

    local out requestComp lastParam lastChar comp directive args

    # Prepare the command to request completions for the program.
    # Calling ${words[0]} instead of directly flightctl allows handling aliases
    args=("${words[@]:1}")
    # Disable ActiveHelp which is not supported for bash completion v1
    requestComp="FLIGHTCTL_ACTIVE_HELP=0 ${words[0]} __completeNoDesc ${args[*]}"

    lastParam=${words[$((${#words[@]}-1))]}
    lastChar=${lastParam:$((${#lastParam}-1)):1}
    __flightctl_debug "${FUNCNAME[0]}: lastParam ${lastParam}, lastChar ${lastChar}"

    if [ -z "${cur}" ] && [ "${lastChar}" != "=" ]; then
        # If the last parameter is complete (there is a space following it)
        # We add an extra empty parameter so we can indicate this to the go method.
        __flightctl_debug "${FUNCNAME[0]}: Adding extra empty parameter"
        requestComp="${requestComp} \"\""
    fi

    __flightctl_debug "${FUNCNAME[0]}: calling ${requestComp}"
    # Use eval to handle any environment variables and such
    out=$(eval "${requestComp}" 2>/dev/null)

    # Extract the directive integer at the very end of the output following a colon (:)
    directive=${out##*:}
    # Remove the directive
    out=${out%:*}
    if [ "${directive}" = "${out}" ]; then
        # There is not directive specified
        directive=0
    fi
    __flightctl_debug "${FUNCNAME[0]}: the completion directive is: ${directive}"
    __flightctl_debug "${FUNCNAME[0]}: the completions are: ${out}"

    if [ $((directive & shellCompDirectiveError)) -ne 0 ]; then
        # Error code.  No completion.
        __flightctl_debug "${FUNCNAME[0]}: received error from custom completion go code"
        return
    else
        if [ $((directive & shellCompDirectiveNoSpace)) -ne 0 ]; then
            if [[ $(type -t compopt) = "builtin" ]]; then
                __flightctl_debug "${FUNCNAME[0]}: activating no space"
                compopt -o nospace
            fi
        fi
        if [ $((directive & shellCompDirectiveNoFileComp)) -ne 0 ]; then
            if [[ $(type -t compopt) = "builtin" ]]; then
                __flightctl_debug "${FUNCNAME[0]}: activating no file completion"
                compopt +o default
            fi
        fi
    fi

    if [ $((directive & shellCompDirectiveFilterFileExt)) -ne 0 ]; then
        # File extension filtering
        local fullFilter filter filteringCmd
        # Do not use quotes around the $out variable or else newline
        # characters will be kept.
        for filter in ${out}; do
            fullFilter+="$filter|"
        done

        filteringCmd="_filedir $fullFilter"
        __flightctl_debug "File filtering command: $filteringCmd"
        $filteringCmd
    elif [ $((directive & shellCompDirectiveFilterDirs)) -ne 0 ]; then
        # File completion for directories only
        local subdir
        # Use printf to strip any trailing newline
        subdir=$(printf "%s" "${out}")
        if [ -n "$subdir" ]; then
            __flightctl_debug "Listing directories in $subdir"
            __flightctl_handle_subdirs_in_dir_flag "$subdir"
        else
            __flightctl_debug "Listing directories in ."
            _filedir -d
        fi
    else
        while IFS='' read -r comp; do
            COMPREPLY+=("$comp")
        done < <(compgen -W "${out}" -- "$cur")
    fi
}

__flightctl_handle_reply()
{
    __flightctl_debug "${FUNCNAME[0]}"
    local comp
    case $cur in
        -*)
            if [[ $(type -t compopt) = "builtin" ]]; then
                compopt -o nospace
            fi
            local allflags
            if [ ${#must_have_one_flag[@]} -ne 0 ]; then
                allflags=("${must_have_one_flag[@]}")
            else
                allflags=("${flags[*]} ${two_word_flags[*]}")
            fi
            while IFS='' read -r comp; do
                COMPREPLY+=("$comp")
            done < <(compgen -W "${allflags[*]}" -- "$cur")
            if [[ $(type -t compopt) = "builtin" ]]; then
                [[ "${COMPREPLY[0]}" == *= ]] || compopt +o nospace
            fi

            # complete after --flag=abc
            if [[ $cur == *=* ]]; then
                if [[ $(type -t compopt) = "builtin" ]]; then
                    compopt +o nospace
                fi

                local index flag
                flag="${cur%=*}"
                __flightctl_index_of_word "${flag}" "${flags_with_completion[@]}"
                COMPREPLY=()
                if [[ ${index} -ge 0 ]]; then
                    PREFIX=""
                    cur="${cur#*=}"
                    ${flags_completion[${index}]}
                    if [ -n "${ZSH_VERSION:-}" ]; then
                        # zsh completion needs --flag= prefix
                        eval "COMPREPLY=( \"\${COMPREPLY[@]/#/${flag}=}\" )"
                    fi
                fi
            fi

            if [[ -z "${flag_parsing_disabled}" ]]; then
                # If flag parsing is enabled, we have completed the flags and can return.
                # If flag parsing is disabled, we may not know all (or any) of the flags, so we fallthrough
                # to possibly call handle_go_custom_completion.
                return 0;
            fi
            ;;
    esac

    # check if we are handling a flag with special work handling
    local index
    __flightctl_index_of_word "${prev}" "${flags_with_completion[@]}"
    if [[ ${index} -ge 0 ]]; then
        ${flags_completion[${index}]}
        return
    fi

    # we are parsing a flag and don't have a special handler, no completion
    if [[ ${cur} != "${words[cword]}" ]]; then
        return
    fi

    local completions
    completions=("${commands[@]}")
    if [[ ${#must_have_one_noun[@]} -ne 0 ]]; then
        completions+=("${must_have_one_noun[@]}")
    elif [[ -n "${has_completion_function}" ]]; then
        # if a go completion function is provided, defer to that function
        __flightctl_handle_go_custom_completion
    fi
    if [[ ${#must_have_one_flag[@]} -ne 0 ]]; then
        completions+=("${must_have_one_flag[@]}")
    fi
    while IFS='' read -r comp; do
        COMPREPLY+=("$comp")
    done < <(compgen -W "${completions[*]}" -- "$cur")

    if [[ ${#COMPREPLY[@]} -eq 0 && ${#noun_aliases[@]} -gt 0 && ${#must_have_one_noun[@]} -ne 0 ]]; then
        while IFS='' read -r comp; do
            COMPREPLY+=("$comp")
        done < <(compgen -W "${noun_aliases[*]}" -- "$cur")
    fi

    if [[ ${#COMPREPLY[@]} -eq 0 ]]; then
        if declare -F __flightctl_custom_func >/dev/null; then
            # try command name qualified custom func
            __flightctl_custom_func
        else
            # otherwise fall back to unqualified for compatibility
            declare -F __custom_func >/dev/null && __custom_func
        fi
    fi

    # available in bash-completion >= 2, not always present on macOS
    if declare -F __ltrim_colon_completions >/dev/null; then
        __ltrim_colon_completions "$cur"
    fi

    # If there is only 1 completion and it is a flag with an = it will be completed
    # but we don't want a space after the =
    if [[ "${#COMPREPLY[@]}" -eq "1" ]] && [[ $(type -t compopt) = "builtin" ]] && [[ "${COMPREPLY[0]}" == --*= ]]; then
       compopt -o nospace
    fi
}

# The arguments should be in the form "ext1|ext2|extn"
__flightctl_handle_filename_extension_flag()
{
    local ext="$1"
    _filedir "@(${ext})"
}

__flightctl_handle_subdirs_in_dir_flag()
{
    local dir="$1"
    pushd "${dir}" >/dev/null 2>&1 && _filedir -d && popd >/dev/null 2>&1 || return
}

__flightctl_handle_flag()
{
    __flightctl_debug "${FUNCNAME[0]}: c is $c words[c] is ${words[c]}"

    # if a command required a flag, and we found it, unset must_have_one_flag()
    local flagname=${words[c]}
    local flagvalue=""
    # if the word contained an =
    if [[ ${words[c]} == *"="* ]]; then
        flagvalue=${flagname#*=} # take in as flagvalue after the =
        flagname=${flagname%=*} # strip everything after the =
        flagname="${flagname}=" # but put the = back
    fi
    __flightctl_debug "${FUNCNAME[0]}: looking for ${flagname}"
    if __flightctl_contains_word "${flagname}" "${must_have_one_flag[@]}"; then
        must_have_one_flag=()
    fi

    # if you set a flag which only applies to this command, don't show subcommands
    if __flightctl_contains_word "${flagname}" "${local_nonpersistent_flags[@]}"; then
      commands=()
    fi

    # keep flag value with flagname as flaghash
    # flaghash variable is an associative array which is only supported in bash > 3.
    if [[ -z "${BASH_VERSION:-}" || "${BASH_VERSINFO[0]:-}" -gt 3 ]]; then
        if [ -n "${flagvalue}" ] ; then
            flaghash[${flagname}]=${flagvalue}
        elif [ -n "${words[ $((c+1)) ]}" ] ; then
            flaghash[${flagname}]=${words[ $((c+1)) ]}
        else
            flaghash[${flagname}]="true" # pad "true" for bool flag
        fi
    fi

    # skip the argument to a two word flag
    if [[ ${words[c]} != *"="* ]] && __flightctl_contains_word "${words[c]}" "${two_word_flags[@]}"; then
        __flightctl_debug "${FUNCNAME[0]}: found a flag ${words[c]}, skip the next argument"
        c=$((c+1))
        # if we are looking for a flags value, don't show commands
        if [[ $c -eq $cword ]]; then
            commands=()
        fi
    fi

    c=$((c+1))

}

__flightctl_handle_noun()
{
    __flightctl_debug "${FUNCNAME[0]}: c is $c words[c] is ${words[c]}"

    if __flightctl_contains_word "${words[c]}" "${must_have_one_noun[@]}"; then
        must_have_one_noun=()
    elif __flightctl_contains_word "${words[c]}" "${noun_aliases[@]}"; then
        must_have_one_noun=()
    fi

    nouns+=("${words[c]}")
    c=$((c+1))
}

__flightctl_handle_command()
{
    __flightctl_debug "${FUNCNAME[0]}: c is $c words[c] is ${words[c]}"

    local next_command
    if [[ -n ${last_command} ]]; then
        next_command="_${last_command}_${words[c]//:/__}"
    else
        if [[ $c -eq 0 ]]; then
            next_command="_flightctl_root_command"
        else
            next_command="_${words[c]//:/__}"
        fi
    fi
    c=$((c+1))
    __flightctl_debug "${FUNCNAME[0]}: looking for ${next_command}"
    declare -F "$next_command" >/dev/null && $next_command
}

__flightctl_handle_word()
{
    if [[ $c -ge $cword ]]; then
        __flightctl_handle_reply
        return
    fi
    __flightctl_debug "${FUNCNAME[0]}: c is $c words[c] is ${words[c]}"
    if [[ "${words[c]}" == -* ]]; then
        __flightctl_handle_flag
    elif __flightctl_contains_word "${words[c]}" "${commands[@]}"; then
        __flightctl_handle_command
    elif [[ $c -eq 0 ]]; then
        __flightctl_handle_command
    elif __flightctl_contains_word "${words[c]}" "${command_aliases[@]}"; then
        # aliashash variable is an associative array which is only supported in bash > 3.
        if [[ -z "${BASH_VERSION:-}" || "${BASH_VERSINFO[0]:-}" -gt 3 ]]; then
            words[c]=${aliashash[${words[c]}]}
            __flightctl_handle_command
        else
            __flightctl_handle_noun
        fi
    else
        __flightctl_handle_noun
    fi
    __flightctl_handle_word
}

_flightctl_apply()
{
    last_command="flightctl_apply"

    command_aliases=()

    commands=()

    flags=()
    two_word_flags=()
    local_nonpersistent_flags=()
    flags_with_completion=()
    flags_completion=()

    flags+=("--context=")
    two_word_flags+=("--context")
    two_word_flags+=("-c")
    local_nonpersistent_flags+=("--context")
    local_nonpersistent_flags+=("--context=")
    local_nonpersistent_flags+=("-c")
    flags+=("--dry-run")
    local_nonpersistent_flags+=("--dry-run")
    flags+=("--filename=")
    two_word_flags+=("--filename")
    flags_with_completion+=("--filename")
    flags_completion+=("__flightctl_handle_filename_extension_flag json|yaml|yml")
    two_word_flags+=("-f")
    flags_with_completion+=("-f")
    flags_completion+=("__flightctl_handle_filename_extension_flag json|yaml|yml")
    local_nonpersistent_flags+=("--filename")
    local_nonpersistent_flags+=("--filename=")
    local_nonpersistent_flags+=("-f")
    flags+=("--recursive")
    flags+=("-R")
    local_nonpersistent_flags+=("--recursive")
    local_nonpersistent_flags+=("-R")

    must_have_one_flag=()
    must_have_one_noun=()
    noun_aliases=()
}

_flightctl_approve()
{
    last_command="flightctl_approve"

    command_aliases=()

    commands=()

    flags=()
    two_word_flags=()
    local_nonpersistent_flags=()
    flags_with_completion=()
    flags_completion=()

    flags+=("--context=")
    two_word_flags+=("--context")
    two_word_flags+=("-c")
    local_nonpersistent_flags+=("--context")
    local_nonpersistent_flags+=("--context=")
    local_nonpersistent_flags+=("-c")
    flags+=("--label=")
    two_word_flags+=("--label")
    two_word_flags+=("-l")
    local_nonpersistent_flags+=("--label")
    local_nonpersistent_flags+=("--label=")
    local_nonpersistent_flags+=("-l")

    must_have_one_flag=()
    must_have_one_noun=()
    noun_aliases=()
}

_flightctl_certificate()
{
    last_command="flightctl_certificate"

    command_aliases=()

    commands=()

    flags=()
    two_word_flags=()
    local_nonpersistent_flags=()
    flags_with_completion=()
    flags_completion=()

    flags+=("--cert-type=")
    two_word_flags+=("--cert-type")
    two_word_flags+=("-t")
    local_nonpersistent_flags+=("--cert-type")
    local_nonpersistent_flags+=("--cert-type=")
    local_nonpersistent_flags+=("-t")
    flags+=("--context=")
    two_word_flags+=("--context")
    two_word_flags+=("-c")
    local_nonpersistent_flags+=("--context")
    local_nonpersistent_flags+=("--context=")
    local_nonpersistent_flags+=("-c")
    flags+=("--encrypt")
    flags+=("-s")
    local_nonpersistent_flags+=("--encrypt")
    local_nonpersistent_flags+=("-s")
    flags+=("--expiration=")
    two_word_flags+=("--expiration")
    two_word_flags+=("-x")
    local_nonpersistent_flags+=("--expiration")
    local_nonpersistent_flags+=("--expiration=")
    local_nonpersistent_flags+=("-x")
    flags+=("--name=")
    two_word_flags+=("--name")
    two_word_flags+=("-n")
    local_nonpersistent_flags+=("--name")
    local_nonpersistent_flags+=("--name=")
    local_nonpersistent_flags+=("-n")
    flags+=("--output-dir=")
    two_word_flags+=("--output-dir")
    two_word_flags+=("-d")
    local_nonpersistent_flags+=("--output-dir")
    local_nonpersistent_flags+=("--output-dir=")
    local_nonpersistent_flags+=("-d")
    flags+=("--output-format=")
    two_word_flags+=("--output-format")
    two_word_flags+=("-o")
    local_nonpersistent_flags+=("--output-format")
    local_nonpersistent_flags+=("--output-format=")
    local_nonpersistent_flags+=("-o")

    must_have_one_flag=()
    must_have_one_noun=()
    must_have_one_noun+=("request")
    noun_aliases=()
}

_flightctl_completion()
{
    last_command="flightctl_completion"

    command_aliases=()

    commands=()

    flags=()
    two_word_flags=()
    local_nonpersistent_flags=()
    flags_with_completion=()
    flags_completion=()

    flags+=("--help")
    flags+=("-h")
    local_nonpersistent_flags+=("--help")
    local_nonpersistent_flags+=("-h")

    must_have_one_flag=()
    must_have_one_noun=()
    must_have_one_noun+=("PowerShell")
    must_have_one_noun+=("bash")
    must_have_one_noun+=("fish")
    must_have_one_noun+=("zsh")
    noun_aliases=()
}

_flightctl_console()
{
    last_command="flightctl_console"

    command_aliases=()

    commands=()

    flags=()
    two_word_flags=()
    local_nonpersistent_flags=()
    flags_with_completion=()
    flags_completion=()

    flags+=("--context=")
    two_word_flags+=("--context")
    two_word_flags+=("-c")
    local_nonpersistent_flags+=("--context")
    local_nonpersistent_flags+=("--context=")
    local_nonpersistent_flags+=("-c")

    must_have_one_flag=()
    must_have_one_noun=()
    noun_aliases=()
}

_flightctl_csr-generate()
{
    last_command="flightctl_csr-generate"

    command_aliases=()

    commands=()

    flags=()
    two_word_flags=()
    local_nonpersistent_flags=()
    flags_with_completion=()
    flags_completion=()

    flags+=("--context=")
    two_word_flags+=("--context")
    two_word_flags+=("-c")
    local_nonpersistent_flags+=("--context")
    local_nonpersistent_flags+=("--context=")
    local_nonpersistent_flags+=("-c")
    flags+=("--expiry=")
    two_word_flags+=("--expiry")
    two_word_flags+=("-e")
    local_nonpersistent_flags+=("--expiry")
    local_nonpersistent_flags+=("--expiry=")
    local_nonpersistent_flags+=("-e")
    flags+=("--name=")
    two_word_flags+=("--name")
    two_word_flags+=("-n")
    local_nonpersistent_flags+=("--name")
    local_nonpersistent_flags+=("--name=")
    local_nonpersistent_flags+=("-n")
    flags+=("--output=")
    two_word_flags+=("--output")
    two_word_flags+=("-o")
    local_nonpersistent_flags+=("--output")
    local_nonpersistent_flags+=("--output=")
    local_nonpersistent_flags+=("-o")
    flags+=("--overwrite")
    flags+=("-y")
    local_nonpersistent_flags+=("--overwrite")
    local_nonpersistent_flags+=("-y")

    must_have_one_flag=()
    must_have_one_flag+=("--output=")
    must_have_one_flag+=("-o")
    must_have_one_noun=()
    noun_aliases=()
}

_flightctl_decommission()
{
    last_command="flightctl_decommission"

    command_aliases=()

    commands=()

    flags=()
    two_word_flags=()
    local_nonpersistent_flags=()
    flags_with_completion=()
    flags_completion=()

    flags+=("--context=")
    two_word_flags+=("--context")
    two_word_flags+=("-c")
    local_nonpersistent_flags+=("--context")
    local_nonpersistent_flags+=("--context=")
    local_nonpersistent_flags+=("-c")
    flags+=("--target=")
    two_word_flags+=("--target")
    two_word_flags+=("-t")
    local_nonpersistent_flags+=("--target")
    local_nonpersistent_flags+=("--target=")
    local_nonpersistent_flags+=("-t")

    must_have_one_flag=()
    must_have_one_noun=()
    noun_aliases=()
}

_flightctl_delete()
{
    last_command="flightctl_delete"

    command_aliases=()

    commands=()

    flags=()
    two_word_flags=()
    local_nonpersistent_flags=()
    flags_with_completion=()
    flags_completion=()

    flags+=("--context=")
    two_word_flags+=("--context")
    two_word_flags+=("-c")
    local_nonpersistent_flags+=("--context")
    local_nonpersistent_flags+=("--context=")
    local_nonpersistent_flags+=("-c")
    flags+=("--fleetname=")
    two_word_flags+=("--fleetname")
    two_word_flags+=("-f")
    local_nonpersistent_flags+=("--fleetname")
    local_nonpersistent_flags+=("--fleetname=")
    local_nonpersistent_flags+=("-f")

    must_have_one_flag=()
    must_have_one_noun=()
    must_have_one_noun+=("certificatesigningrequests")
    must_have_one_noun+=("devices")
    must_have_one_noun+=("enrollmentrequests")
    must_have_one_noun+=("fleets")
    must_have_one_noun+=("repositories")
    must_have_one_noun+=("resourcesyncs")
    must_have_one_noun+=("templateversions")
    noun_aliases=()
}

_flightctl_deny()
{
    last_command="flightctl_deny"

    command_aliases=()

    commands=()

    flags=()
    two_word_flags=()
    local_nonpersistent_flags=()
    flags_with_completion=()
    flags_completion=()

    flags+=("--context=")
    two_word_flags+=("--context")
    two_word_flags+=("-c")
    local_nonpersistent_flags+=("--context")
    local_nonpersistent_flags+=("--context=")
    local_nonpersistent_flags+=("-c")

    must_have_one_flag=()
    must_have_one_noun=()
    noun_aliases=()
}

_flightctl_enrollmentconfig()
{
    last_command="flightctl_enrollmentconfig"

    command_aliases=()

    commands=()

    flags=()
    two_word_flags=()
    local_nonpersistent_flags=()
    flags_with_completion=()
    flags_completion=()

    flags+=("--context=")
    two_word_flags+=("--context")
    two_word_flags+=("-c")
    local_nonpersistent_flags+=("--context")
    local_nonpersistent_flags+=("--context=")
    local_nonpersistent_flags+=("-c")
    flags+=("--private-key=")
    two_word_flags+=("--private-key")
    two_word_flags+=("-p")
    local_nonpersistent_flags+=("--private-key")
    local_nonpersistent_flags+=("--private-key=")
    local_nonpersistent_flags+=("-p")

    must_have_one_flag=()
    must_have_one_noun=()
    noun_aliases=()
}

_flightctl_get()
{
    last_command="flightctl_get"

    command_aliases=()

    commands=()

    flags=()
    two_word_flags=()
    local_nonpersistent_flags=()
    flags_with_completion=()
    flags_completion=()

    flags+=("--context=")
    two_word_flags+=("--context")
    two_word_flags+=("-c")
    local_nonpersistent_flags+=("--context")
    local_nonpersistent_flags+=("--context=")
    local_nonpersistent_flags+=("-c")
    flags+=("--continue=")
    two_word_flags+=("--continue")
    local_nonpersistent_flags+=("--continue")
    local_nonpersistent_flags+=("--continue=")
    flags+=("--field-selector=")
    two_word_flags+=("--field-selector")
    local_nonpersistent_flags+=("--field-selector")
    local_nonpersistent_flags+=("--field-selector=")
    flags+=("--fleetname=")
    two_word_flags+=("--fleetname")
    local_nonpersistent_flags+=("--fleetname")
    local_nonpersistent_flags+=("--fleetname=")
    flags+=("--limit=")
    two_word_flags+=("--limit")
    local_nonpersistent_flags+=("--limit")
    local_nonpersistent_flags+=("--limit=")
    flags+=("--output=")
    two_word_flags+=("--output")
    two_word_flags+=("-o")
    local_nonpersistent_flags+=("--output")
    local_nonpersistent_flags+=("--output=")
    local_nonpersistent_flags+=("-o")
    flags+=("--rendered")
    local_nonpersistent_flags+=("--rendered")
    flags+=("--selector=")
    two_word_flags+=("--selector")
    two_word_flags+=("-l")
    local_nonpersistent_flags+=("--selector")
    local_nonpersistent_flags+=("--selector=")
    local_nonpersistent_flags+=("-l")
    flags+=("--summary")
    flags+=("-s")
    local_nonpersistent_flags+=("--summary")
    local_nonpersistent_flags+=("-s")
    flags+=("--summary-only")
    local_nonpersistent_flags+=("--summary-only")

    must_have_one_flag=()
    must_have_one_noun=()
    must_have_one_noun+=("certificatesigningrequests")
    must_have_one_noun+=("devices")
    must_have_one_noun+=("enrollmentrequests")
    must_have_one_noun+=("fleets")
    must_have_one_noun+=("repositories")
    must_have_one_noun+=("resourcesyncs")
    must_have_one_noun+=("templateversions")
    noun_aliases=()
}

_flightctl_help()
{
    last_command="flightctl_help"

    command_aliases=()

    commands=()

    flags=()
    two_word_flags=()
    local_nonpersistent_flags=()
    flags_with_completion=()
    flags_completion=()


    must_have_one_flag=()
    must_have_one_noun=()
    has_completion_function=1
    noun_aliases=()
}

_flightctl_login()
{
    last_command="flightctl_login"

    command_aliases=()

    commands=()

    flags=()
    two_word_flags=()
    local_nonpersistent_flags=()
    flags_with_completion=()
    flags_completion=()

    flags+=("--auth-certificate-authority=")
    two_word_flags+=("--auth-certificate-authority")
    local_nonpersistent_flags+=("--auth-certificate-authority")
    local_nonpersistent_flags+=("--auth-certificate-authority=")
    flags+=("--certificate-authority=")
    two_word_flags+=("--certificate-authority")
    local_nonpersistent_flags+=("--certificate-authority")
    local_nonpersistent_flags+=("--certificate-authority=")
    flags+=("--client-id=")
    two_word_flags+=("--client-id")
    local_nonpersistent_flags+=("--client-id")
    local_nonpersistent_flags+=("--client-id=")
    flags+=("--context=")
    two_word_flags+=("--context")
    two_word_flags+=("-c")
    local_nonpersistent_flags+=("--context")
    local_nonpersistent_flags+=("--context=")
    local_nonpersistent_flags+=("-c")
    flags+=("--insecure-skip-tls-verify")
    flags+=("-k")
    local_nonpersistent_flags+=("--insecure-skip-tls-verify")
    local_nonpersistent_flags+=("-k")
    flags+=("--token=")
    two_word_flags+=("--token")
    two_word_flags+=("-t")
    local_nonpersistent_flags+=("--token")
    local_nonpersistent_flags+=("--token=")
    local_nonpersistent_flags+=("-t")
    flags+=("--web")
    flags+=("-w")
    local_nonpersistent_flags+=("--web")
    local_nonpersistent_flags+=("-w")

    must_have_one_flag=()
    must_have_one_noun=()
    noun_aliases=()
}

_flightctl_version()
{
    last_command="flightctl_version"

    command_aliases=()

    commands=()

    flags=()
    two_word_flags=()
    local_nonpersistent_flags=()
    flags_with_completion=()
    flags_completion=()

    flags+=("--output=")
    two_word_flags+=("--output")
    two_word_flags+=("-o")
    local_nonpersistent_flags+=("--output")
    local_nonpersistent_flags+=("--output=")
    local_nonpersistent_flags+=("-o")

    must_have_one_flag=()
    must_have_one_noun=()
    noun_aliases=()
}

_flightctl_root_command()
{
    last_command="flightctl"

    command_aliases=()

    commands=()
    commands+=("apply")
    commands+=("approve")
    commands+=("certificate")
    commands+=("completion")
    commands+=("console")
    commands+=("csr-generate")
    commands+=("decommission")
    commands+=("delete")
    commands+=("deny")
    commands+=("enrollmentconfig")
    commands+=("get")
    commands+=("help")
    commands+=("login")
    commands+=("version")

    flags=()
    two_word_flags=()
    local_nonpersistent_flags=()
    flags_with_completion=()
    flags_completion=()


    must_have_one_flag=()
    must_have_one_noun=()
    noun_aliases=()
}

__start_flightctl()
{
    local cur prev words cword split
    declare -A flaghash 2>/dev/null || :
    declare -A aliashash 2>/dev/null || :
    if declare -F _init_completion >/dev/null 2>&1; then
        _init_completion -s || return
    else
        __flightctl_init_completion -n "=" || return
    fi

    local c=0
    local flag_parsing_disabled=
    local flags=()
    local two_word_flags=()
    local local_nonpersistent_flags=()
    local flags_with_completion=()
    local flags_completion=()
    local commands=("flightctl")
    local command_aliases=()
    local must_have_one_flag=()
    local must_have_one_noun=()
    local has_completion_function=""
    local last_command=""
    local nouns=()
    local noun_aliases=()

    __flightctl_handle_word
}

if [[ $(type -t compopt) = "builtin" ]]; then
    complete -o default -F __start_flightctl flightctl
else
    complete -o default -o nospace -F __start_flightctl flightctl
fi

# ex: ts=4 sw=4 et filetype=sh
